import numpy as np
import argparse
import json
import kerrgeopy as kg
from . import trajectory as pn

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--params', required=True, help='JSON file with EMRI/MBH parameters')
    parser.add_argument('--windows', required=True, help='Observation windows file')
    parser.add_argument('--output-timings', required=True)
    parser.add_argument('--dt', type=float, default=10.0)
    args = parser.parse_args()
    
    with open(args.params) as f:
        p = json.load(f)
    
    windows = np.loadtxt(args.windows, ndmin=2)    
    t_g = 4.926580927874239e-06 * 10**p['logMbh']

    _, _, _, lambd, P_orb = pn.trajectory(
        windows,
        np.array([p['logMbh']]), np.array([p['sma']]), np.array([p['e']]),
        np.array([p['incl']]), np.array([p['spin']]), np.array([p['phi_r0']]),
        np.array([p['phi_theta0']]), np.array([p['phi_phi0']]), args.dt
    )
    P_d = p['P_d'] * P_orb[0]
    lambd = lambd[0]
    
    orbit = kg.StableOrbit(
        p['spin'],
        p['sma'] * (1 - p['e']**2),
        p['e'],
        np.cos(np.radians(p['incl']))
    )
    tgeo_func, rgeo_func, thetageo_func, phigeo_func = orbit.trajectory(
        (0, np.pi + p['phi_r0'], -np.pi/2 + p['phi_theta0'], p['phi_phi0'])
    )
    
    tgeo = tgeo_func(lambd) * t_g
    rgeo = rgeo_func(lambd)
    thetageo = thetageo_func(lambd)
    phigeo = phigeo_func(lambd)
    
    sin_theta = np.sin(thetageo)
    xgeo = rgeo * sin_theta * np.cos(phigeo)
    ygeo = rgeo * sin_theta * np.sin(phigeo)
    zgeo = rgeo * np.cos(thetageo)
    
    theta_d_rad = np.radians(p['theta_d'])
    disk_phase = 2 * np.pi * tgeo / P_d + p['phi_d']
    n_crs_x = np.sin(theta_d_rad) * np.cos(disk_phase)
    n_crs_y = np.sin(theta_d_rad) * np.sin(disk_phase)
    n_crs_z = np.cos(theta_d_rad)
    
    D_t = n_crs_x * xgeo + n_crs_y * ygeo + n_crs_z * zgeo
    
    crossings_mask = (D_t[:-1] * D_t[1:]) < 0
    alpha = -D_t[:-1] / (D_t[1:] - D_t[:-1])
    
    t_cross = tgeo[:-1] + alpha * (tgeo[1:] - tgeo[:-1])
    x_cross = xgeo[:-1] + alpha * (xgeo[1:] - xgeo[:-1])
    y_cross = ygeo[:-1] + alpha * (ygeo[1:] - ygeo[:-1])
    z_cross = zgeo[:-1] + alpha * (zgeo[1:] - zgeo[:-1])
    r_cross = rgeo[:-1] + alpha * (rgeo[1:] - rgeo[:-1])
    
    r_mag = np.sqrt(x_cross**2 + y_cross**2 + z_cross**2)
    cos_angle = (np.sin(p['theta_obs']) * x_cross + np.cos(p['theta_obs']) * z_cross) / r_mag
    
    shapiro_delay = -2 * t_g * np.log(r_cross * (1 - cos_angle))
    geometric_delay = r_cross * cos_angle * t_g
    
    crossings = np.where(crossings_mask, t_cross + shapiro_delay + geometric_delay, np.nan)
    np.savetxt(args.output_timings, crossings[~np.isnan(crossings)])

if __name__ == '__main__':
    main()


import numpy as np
import argparse
import json
import ultranest
import matplotlib.pyplot as plt
import corner
from ultranest.popstepsampler import PopulationSimpleSliceSampler, PopulationSliceSampler, PopulationRandomWalkSampler, generate_region_oriented_direction, generate_random_direction, generate_mixture_random_direction, generate_cube_oriented_direction
from . import trajectory as pn

DIRECTION_FUNCS = {
    'region': generate_region_oriented_direction,
    'random': generate_random_direction,
    'mixture': generate_mixture_random_direction,
    'cube': generate_cube_oriented_direction,
}

REGION_CLASSES = {
    'mlfriends': ultranest.mlfriends.MLFriends,
    'simple': ultranest.mlfriends.SimpleRegion,
    'ellipsoid': ultranest.mlfriends.RobustEllipsoidRegion,
}

SAMPLERS = {
    'none': None,
    'slice': PopulationSimpleSliceSampler,
    'harm': PopulationSliceSampler,
    'rwalk': PopulationRandomWalkSampler,
}

def load_priors(path):
    with open(path) as f:
        return json.load(f)

def make_prior_transform(param_lims):
    lims = np.array(list(param_lims.values()))
    def transform(cube):
        cube = np.atleast_2d(cube)
        if cube.shape[1] != len(lims) and cube.shape[0] == len(lims):
            cube = cube.T
        return cube * (lims[:, 1] - lims[:, 0]) + lims[:, 0]
    return transform

def make_log_likelihood(timings, windows, errs, dt, one_crossing):
    def loglike(params):
        params = np.atleast_2d(params).astype(np.float32)
        resid = pn.residuals(timings, windows, errs, *params.T, dt, one_crossing)
        ll = -0.5 * pn.to_numpy(resid)
        return np.where(np.isfinite(ll), ll, -1e30).astype(np.float64)
    return loglike

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--output', default='output', help='Sampling output directory')
    parser.add_argument('--timings', default='timings.txt', help='.txt file containing QPE timings (one per line)')
    parser.add_argument('--windows', default='windows.txt', help='.txt file containing observation windows (one start-stop pair per line separated by a space)')
    parser.add_argument('--errors', default='errors.txt', help='.txt file containing QPE timing errors (one per line)')
    parser.add_argument('--priors', default='priors.json', help='.json file containing sampling priors')
    parser.add_argument('--dt', type=float, default=10.0, help='Time step size for likelihood evaluations')
    parser.add_argument('--gpu', action='store_true', help='Use GPU-accelerated likelihood evaluation')
    parser.add_argument('--stepsampler', choices=SAMPLERS.keys(), default='slice', help='Which step sampler to use. Options are none, slice, harm, rwalk.')
    parser.add_argument('--direction', choices=DIRECTION_FUNCS.keys(), default='region', help='Which direction function to use for step sampling. Options are region, random, mixture, cube.')
    parser.add_argument('--region', choices=REGION_CLASSES.keys(), default='simple', help='Which region class to use. Options are mlfriends, simple, ellipsoid.')
    parser.add_argument('--popsize', type=int, default=256, help='Number of walkers to maintain for step sampling.')
    parser.add_argument('--nsteps', type=int, default=256, help='Number of steps for step sampling.')
    parser.add_argument('--nlive', type=int, default=600, help='Minimum number of live points throughout the run')
    parser.add_argument('--dkl', type=float, default=0.5, help='Target posterior uncertainty (KL divergence b/w bootstrapped integrators, in nats)')
    parser.add_argument('--frac-remain', type=float, default=0.01, help='Integrate until this fraction of the integral is left in the remainder.')
    parser.add_argument('--min-ess', type=int, default=400, help='Minimum effective sample size for nested sampling.')
    parser.add_argument('--one-crossing', action='store_true', help='Keep only one crossing per orbit.')
    args = parser.parse_args()

    pn.set_backend(args.gpu)
    
    timings = pn.xp.asarray(np.loadtxt(args.timings), dtype=pn.xp.float32)
    windows = np.loadtxt(args.windows, ndmin=2)
    errs = pn.xp.asarray(np.loadtxt(args.errors), dtype=pn.xp.float32)
    priors = load_priors(args.priors)
    
    param_names = list(priors.keys())
    wrapped = [priors[p].get('wrapped', False) for p in param_names]
    param_lims = {p: priors[p]['bounds'] for p in param_names}
    
    sampler = ultranest.ReactiveNestedSampler(
        param_names,
        make_log_likelihood(timings, windows, errs, args.dt, args.one_crossing),
        make_prior_transform(param_lims),
        log_dir=args.output,
        resume='resume',
        vectorized=True,
        wrapped_params=wrapped,
    )
    
    if SAMPLERS[args.stepsampler]:
        sampler.stepsampler = SAMPLERS[args.stepsampler](
            popsize=args.popsize,
            nsteps=args.nsteps,
            generate_direction=DIRECTION_FUNCS[args.direction]        )
    
    sampler.run(
        dKL=args.dkl,
        min_num_live_points=args.nlive,
        frac_remain=args.frac_remain,
        min_ess=args.min_ess,
        region_class=REGION_CLASSES[args.region],
    )
    
    sampler.print_results()
    
    fig = corner.corner(
        sampler.results['samples'],
        labels=param_names,
        show_titles=True,
        title_fmt='.3f',
        quantiles=[0.16, 0.5, 0.84],
        bins=50
    )
    fig.savefig(f"{args.output}/corner.png", dpi=150, bbox_inches='tight')
    plt.close(fig)

if __name__ == '__main__':
    main()